@@ -39,6 +39,8 @@ var valid_ds_extensions = {
   '.kml':'ogr',
   '.rss':'ogr',
   '.gpx':'ogr',
+  '.gdb':'ogr',
+  '.topojson':'ogr',
   '.db':'sqlite',
   '.sqlite3':'sqlite',
   '.sqlite':'sqlite',
@@ -72,7 +74,7 @@ var SRS = {
 // on object of locks for concurrent downloads
 var downloads = {};
 // object for tracking logging on downloads
-var download_log_interval = true;
+var download_log_interval = null;
 // last download count, in order to limit logging barf
 var download_last_count = 0;
 
@@ -89,20 +91,23 @@ var pool = require('generic-pool').Pool({
 
 function download(url, options, callback) {
     if (env == 'development') {
-        download_log_interval = setInterval(function() {
-            var in_use = Object.keys(downloads);
-            if (in_use.length > 0 && (download_last_count !== in_use.length)) {
-                var msg = "[millstone] currently downloading " + in_use.length + ' file';
-                if (in_use.length > 1) {
-                    msg += 's';
+        if (!download_log_interval) {
+            download_log_interval = setInterval(function() {
+                var in_use = Object.keys(downloads);
+                if (in_use.length > 0 && (download_last_count !== in_use.length)) {
+                    var msg = "[millstone] currently downloading " + in_use.length + ' file';
+                    if (in_use.length > 1) {
+                        msg += 's';
+                    }
+                    msg += ': ' + _(in_use).map(function(f) { return path.basename(f); });
+                    console.error(msg);
                 }
-                msg += ': ' + _(in_use).map(function(f) { return path.basename(f); });
-                console.error(msg);
-            }
-            download_last_count = in_use.length;
-        },5000);
+                download_last_count = in_use.length;
+            },5000);
+        }
     } else {
         clearInterval(download_log_interval);
+        download_log_interval = null;
     }
     // https://github.com/mapbox/millstone/issues/39
     url = unescape(url);
@@ -270,8 +275,9 @@ function add_item_to_metafile(metafile,key,item,callback) {
 	});
 }
 
-function isRelative(loc) {
-    if (process.platform === 'win32') {
+function isRelative(loc,platform) {
+    platform = platform || process.platform;
+    if (platform === 'win32') {
         return loc[0] !== '\\' && loc[0] !== '/' && loc.match(/^[a-zA-Z]:/) === null;
     } else {
         return loc[0] !== '/';
@@ -437,26 +443,17 @@ function unzip(file, callback) {
                 path.basename(file, path.extname(file)) +
                 path.extname(name).toLowerCase()
             );
-            zf.readFile(name, function(err, buff) {
+            zf.copyFile(name, dest, function(err) {
                 if (err) return callback(err);
-                fs.open(dest, 'w', 0644, function(err, fd) {
-                    if (err) return callback(err);
-                    fs.write(fd, buff, 0, buff.length, null, function(err) {
-                        if (err) return callback(err);
-                        fs.close(fd, function(err) {
-                            if (err) return callback(err);
-                            remain--;
-                            if (!remain) {
-                                add_item_to_metafile(metafile,'unzipped_file',new_name,function(err) {
-                                    // ignore error from add_item_to_metafile
-                                    //if (err && env == 'development') console.error('[millstone] ' + err.message);
-                                    if (err) throw err;
-                                    callback(null, new_name);
-                                });
-                            }
-                        });
+                remain--;
+                if (!remain) {
+                    add_item_to_metafile(metafile,'unzipped_file',new_name,function(err) {
+                        // ignore error from add_item_to_metafile
+                        //if (err && env == 'development') console.error('[millstone] ' + err.message);
+                        if (err) throw err;
+                        callback(null, new_name);
                     });
-                });
+                }
             });
         }
     });
@@ -520,6 +517,7 @@ function resolve(options, callback) {
     if (!options.base) return callback(new Error('options.base is required'));
     if (!options.cache) return callback(new Error('options.cache is required'));
     if (typeof options.nosymlink === "undefined") options.nosymlink = false;
+    if (typeof options.benchmark === "undefined") options.benchmark = false;
     // respect global no-symlinking preference on windows
     if (never_link) options.nosymlink = true;
 
@@ -527,15 +525,16 @@ function resolve(options, callback) {
         base = path.resolve(options.base),
         cache = path.resolve(options.cache),
         resolved = JSON.parse(JSON.stringify(mml)),
-        nosymlink = options.nosymlink;
+        nosymlink = options.nosymlink,
+        benchmark = options.benchmark;
 
     Step(function setup() {
         if (nosymlink) mkdirp(base, 0755, this);
         else mkdirp(path.join(base, 'layers'), 0755, this);
-    }, function externals(err) {
+    }, function style_externals(err) {
         if (err && err.code !== 'EEXIST') throw err;
-
-        var remaining = mml.Layer.length + mml.Stylesheet.length;
+        if (benchmark) console.time("[millstone][benchmark] Resolving style (mss) externals");
+        var remaining = mml.Stylesheet.length;
         var error = null;
         var next = function(err) {
             remaining--;
@@ -545,6 +544,67 @@ function resolve(options, callback) {
 
         if (!remaining) return this();
 
+        // Handle URIs within the Carto CSS
+        function localizeCartoURIs(s,cb) {
+            // Get all unique URIs in stylesheet
+            // TODO - avoid finding non url( uris?
+            var matches = s.data.match(/[-a-zA-Z0-9@:%_\+.~#?&//=]{2,256}\.[a-z]{2,4}\b(\/[-a-zA-Z0-9@:%_\+.~#?&//=]*)?/gi);
+            var URIs = _.uniq(matches || []);
+            var CartoURIs = [];
+            // remove any matched urls that are not clearly
+            // part of a carto style
+            // TODO - consider moving to carto so we can also avoid
+            // downloading commented out code
+            URIs.forEach(function(u) {
+                var idx = s.data.indexOf(u);
+                if (idx > -1) {
+                    var pre_url = s.data.slice(idx-5,idx);
+                    if (pre_url.indexOf('url(') > -1) {
+                        CartoURIs.push(u);
+                        // Update number of async calls so that we don't
+                        // call this() too soon (before everything is done)
+                        remaining += 1;
+                    }
+                }
+            });
+
+            CartoURIs.forEach(function(u) {
+                var uri = url.parse(encodeURI(u));
+                // URL.
+                if (uri.protocol && (uri.protocol == 'http:' || uri.protocol == 'https:')) {
+                    var filepath = path.join(cache, cachepath(u));
+                    localize(uri.href, {filepath:filepath,name:s.id}, function(err, file) {
+                        if (err) {
+                            cb(err);
+                        } else {
+                            var extname = path.extname(file);
+                            if (!extname) {
+                                readExtension(file, function(error, ext) {
+                                    // note - we ignore any readExtension errors
+                                    if (ext) {
+                                        var new_filename = file + ext;
+                                        fs.rename(file, new_filename, function(err) {
+                                            s.data = s.data.split(u).join(new_filename);
+                                            cb(err);
+                                        });
+                                    } else {
+                                        s.data = s.data.split(u).join(file);
+                                        cb(err);
+                                    }
+                                });
+                            } else {
+                                s.data = s.data.split(u).join(file);
+                                cb(err);
+                            }
+                        }
+                    });
+                } else {
+                    cb();
+                }
+            });
+            cb();
+        }
+
         resolved.Stylesheet.forEach(function(s, index) {
             if (typeof s !== 'string') {
                 if (env == 'development') console.error("[millstone] processing style '" + s.id + "'");
@@ -582,70 +642,20 @@ function resolve(options, callback) {
                 localizeCartoURIs(resolved.Stylesheet[index],next);
             });
         });
-
-        // Handle URIs within the Carto CSS
-        function localizeCartoURIs(s,cb) {
-
-            // Get all unique URIs in stylesheet
-            // TODO - avoid finding non url( uris?
-            var matches = s.data.match(/[-a-zA-Z0-9@:%_\+.~#?&//=]{2,256}\.[a-z]{2,4}\b(\/[-a-zA-Z0-9@:%_\+.~#?&//=]*)?/gi);
-            var URIs = _.uniq(matches || []);
-            var CartoURIs = [];
-            // remove any matched urls that are not clearly
-            // part of a carto style
-            // TODO - consider moving to carto so we can also avoid
-            // downloading commented out code
-            URIs.forEach(function(u) {
-                var idx = s.data.indexOf(u);
-                if (idx > -1) {
-                    var pre_url = s.data.slice(idx-5,idx);
-                    if (pre_url.indexOf('url(') > -1) {
-                        CartoURIs.push(u);
-                        // Update number of async calls so that we don't
-                        // call this() too soon (before everything is done)
-                        remaining += 1;
-                    }
-                }
-            });
-
-            CartoURIs.forEach(function(u) {
-                var uri = url.parse(encodeURI(u));
-
-                // URL.
-                if (uri.protocol && (uri.protocol == 'http:' || uri.protocol == 'https:')) {
-                    var filepath = path.join(cache, cachepath(u));
-                    localize(uri.href, {filepath:filepath,name:s.id}, function(err, file) {
-                        if (err) {
-                            cb(err);
-                        } else {
-                            var extname = path.extname(file);
-                            if (!extname) {
-                                readExtension(file, function(error, ext) {
-                                    // note - we ignore any readExtension errors
-                                    if (ext) {
-                                        var new_filename = file + ext;
-                                        fs.rename(file, new_filename, function(err) {
-                                            s.data = s.data.split(u).join(new_filename);
-                                            cb(err);
-                                        });
-                                    } else {
-                                        s.data = s.data.split(u).join(file);
-                                        cb(err);
-                                    }
-                                });
-                            } else {
-                                s.data = s.data.split(u).join(file);
-                                cb(err);
-                            }
-                        }
-                    });
-                } else {
-                    cb();
-                }
-            });
-            cb();
+    }, function layer_externals(err) {
+        if (err) throw err;
+        if (benchmark) {
+            console.timeEnd("[millstone][benchmark] Resolving style (mss) externals");
+            console.time("[millstone][benchmark] Resolving layer (mml) externals");
         }
-
+        var remaining = mml.Layer.length;
+        var error = null;
+        var next = function(err) {
+            remaining--;
+            if (err && err.code !== 'EEXIST') error = err;
+            if (remaining <= 0) this(error);
+        }.bind(this);
+        if (!remaining) return this();
         resolved.Layer.forEach(function(l, index) {
             if (!l.Datasource || !l.Datasource.file) return next();
             if (env == 'development') console.error("[millstone] processing layer '" + l.name + "'");
@@ -763,6 +773,7 @@ function resolve(options, callback) {
         });
     }, function processSql(err) {
         if (err) throw err;
+        if (benchmark) console.timeEnd("[millstone][benchmark] Resolving layer (mml) externals");
         var group = this.group();
         resolved.Layer.forEach(function(l, index) {
             var d = l.Datasource;
@@ -829,7 +840,7 @@ function resolve(options, callback) {
         });
     }, function autodetect(err) {
         if (err) throw err;
-
+        if (benchmark) console.time("[millstone][benchmark] detecting layer type and srs");
         var group = this.group();
         resolved.Layer.forEach(function(l, index) {
             var d = l.Datasource;
@@ -893,7 +904,6 @@ function resolve(options, callback) {
                                 var vrt_file = fs.readFileSync(d.file, 'utf8').toString();
                                 if (vrt_file.indexOf('OGRVRTLayer') != -1) {
                                     d.type = 'ogr';
-                                    d.layer_by_index = 0;
                                     if (!l.srs) {
                                         var match = vrt_file.match(/<LayerSRS>(.+)<\/LayerSRS>/);
                                         if (match && match[1]) {
@@ -914,20 +924,23 @@ function resolve(options, callback) {
                             l.srs = l.srs || SRS.WGS84;
                             break;
                         case '.geojson':
+                        case '.topojson':
                         case '.json':
-                            d.layer_by_index = 0;
                             try {
-                                l.srs = l.srs || srs.parse(d.file).proj4;
+                                var json_obj = JSON.parse(fs.readFileSync(d.file));
+                                l.srs = l.srs || srs.parse(srs.jsonCrs(json_obj)).proj4;
                             } catch (e) {
                                 next(new Error("Could not parse: '" + d.file + "': error: '" + e.message + "'"));
                             }
                             break;
                         case '.kml':
                         case '.rss':
-                            d.layer_by_index = 0;
                             l.srs = SRS.WGS84;
                             break;
                         }
+                        if (d.type === 'ogr' && !d.layer) {
+                            d.layer_by_index = 0;
+                        }
                         // at this point if we do not know the 'type' of mapnik
                         // plugin to dispatch to we are out of luck and there is no
                         // need to check for the projection
@@ -965,12 +978,16 @@ function resolve(options, callback) {
             });
         });
     }, function end(err) {
+        if (benchmark) console.timeEnd("[millstone][benchmark] detecting layer type and srs");
         // Fix map & layer SRS values.
         resolved.srs = resolved.srs || SRS['900913'];
         fixSRS(resolved);
         resolved.Layer.forEach(fixSRS);
         if (!err && env == 'development') console.error("[millstone] finished processing '" + options.base + "'");
-        if (Object.keys(downloads).length < 1) clearInterval(download_log_interval);
+        if (Object.keys(downloads).length < 1) {
+            clearInterval(download_log_interval);
+            download_log_interval = null;
+        }
         return callback(err, resolved);
     });
 }
@@ -1029,11 +1046,13 @@ function flush(options, callback) {
 }
 
 function drainPool(callback) {
-    clearInterval(download_log_interval);
+    if (download_log_interval) clearInterval(download_log_interval);
     if (pool) {
-        pool.drain(function() {
-            pool.destroyAllNow(function() {
-                return callback();
+        pool.drain(function(err) {
+            if (err) return callback(err);
+            pool.destroyAllNow(function(err) {
+               if (err) return callback(err);
+               return callback();
             });
         });
     }
